/**
 * Emanuele Rodola
 * USI Lugano, Nov 2016
 */

#include "mex.h"
#include <vector>
#include <cmath>

//DEBUG
#include <fstream>
#include <iostream>

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
	if (nrhs != 7 || nlhs > 1)
      mexErrMsgTxt("Usage: S = calc_sim_matrix_qap(candidate_matches, DS, DM, sigma, Sdescr, Mdescr, min_dist).");
  
	// NOTE: the code assumes that candidate_matches contains <scene,model> pairs.

	const double* const candidate_matches = mxGetPr(prhs[0]);
	
	if (int(mxGetM(prhs[0])) != 2)
		mexErrMsgTxt("It seems like you only have two candidate matches. Please try to transpose the input matrix");
	
	const int n_candidate_matches = int(mxGetN(prhs[0]));
	mexPrintf("%d candidate matches in total.\n", n_candidate_matches);
	
	const double* const DS = mxGetPr(prhs[1]);
	const double* const DM = mxGetPr(prhs[2]);
	const int n_scene_samples = int(mxGetM(prhs[1]));
	const int n_model_samples = int(mxGetM(prhs[2]));
	
	mexPrintf("%d scene samples.\n", n_scene_samples);
	mexPrintf("%d model samples.\n", n_model_samples);
	
	const double* const Sdescr = mxGetPr(prhs[4]);
	const double* const Mdescr = mxGetPr(prhs[5]);
	const int dim_descr = mxGetM(prhs[4]);
	
	mexPrintf("Descriptors have %d dimensions.\n", dim_descr);
	
	if ( mxGetM(prhs[4]) != mxGetM(prhs[5]) )
		mexErrMsgTxt("The input descriptors must have the same dimensionality.");
	
	if ( mxGetN(prhs[4]) != n_scene_samples )
		mexErrMsgTxt("Please provide a descriptor for each scene point.");
	
	if ( mxGetN(prhs[5]) != n_model_samples )
		mexErrMsgTxt("Please provide a descriptor for each model point.");
	
	if (int(mxGetM(prhs[1])) != int(mxGetN(prhs[1])))
		mexErrMsgTxt("The input distance matrices must be square.");
	
	if (int(mxGetM(prhs[2])) != int(mxGetN(prhs[2])))
		mexErrMsgTxt("The input distance matrices must be square.");
	
	plhs[0] = mxCreateDoubleMatrix(n_candidate_matches, n_candidate_matches, mxREAL);
	double* S = mxGetPr(plhs[0]);
	
	// Note: mxCreateDoubleMatrix() initializes S with zeroes.
	
	const double sigma = *mxGetPr(prhs[3]);
	const double var = 2.*sigma*sigma;
	const double c = 1./(sigma*2.50662827463);
	
	const double min_dist = *mxGetPr(prhs[6]);
	
	mexPrintf("Variance: %.2e\n", var*0.5);
	mexPrintf("Minimum distance among matches: %.3f\n", min_dist);
	
	for (int i=0; i<n_candidate_matches; ++i)
	{
		const int s1 = static_cast<int>( candidate_matches[i*2] ) - 1;   // scene
		const int m1 = static_cast<int>( candidate_matches[i*2+1] ) - 1; // model
		
		for (int j=i; j<n_candidate_matches; ++j)
		{
			const int s2 = static_cast<int>( candidate_matches[j*2] ) - 1;   // scene
			const int m2 = static_cast<int>( candidate_matches[j*2+1] ) - 1; // model
			
			//mexPrintf("(%d , %d) vs (%d , %d)\n", s1, m1, s2, m2);
			
			double sim = 0.;
			
			// prohibit one-to-many and put zero on the diagonal
			if (i==j || s1==s2 || m1==m2)
			{
				S[i*n_candidate_matches+j] = S[j*n_candidate_matches+i] = sim;
				continue;
			}
			
			const double dS = DS[s1*n_scene_samples+s2];
			const double dM = DM[m1*n_model_samples+m2];
			
			// prohibit small distances
			if (dS<min_dist || dM<min_dist)
			{
				S[i*n_candidate_matches+j] = S[j*n_candidate_matches+i] = sim;
				continue;
			}
			
			/*if ((i+1)==113 && (j+1)==745)
			{
				char buf[256];
				sprintf(buf, "%.4f\n", DM[i*n_candidate_matches+j]);
				mexErrMsgTxt(buf);
			}*/
			
			/*double dsim1 = 0., dsim2 = 0.;
			for (int k=0; k<dim_descr; ++k)
			{
				double dd;
				dd = Sdescr[s1*dim_descr+k] - Mdescr[m1*dim_descr+k];
				dsim1 += dd*dd;
				dd = Sdescr[s2*dim_descr+k] - Mdescr[m2*dim_descr+k];
				dsim2 += dd*dd;
			}
			dsim1 = std::exp(-1e-1*dsim1/dim_descr);
			dsim2 = std::exp(-1e-1*dsim2/dim_descr);*/

			sim = dS - dM;
			//mexPrintf("(%d,%d) vs (%d,%d) = %.4f, %.4f, %.4f\n", s1+1, m1+1, s2+1, m2+1, dS, dM, sim);
			//sim *= sim * (dsim1+dsim2);
			sim *= sim;
			//sim = c*std::exp(-sim/var);
			sim = std::exp(-sigma*sim);
			
			S[i*n_candidate_matches+j] = S[j*n_candidate_matches+i] = sim;
		}
	}
	
	return;
}
