/*
 * This file implements the method described in:
 *
 * "Consistent Partial Matching of Shape Collections via Sparse Modeling"
 * L. Cosmo, E. Rodola, A. Albarelli, F. Memoli, D. Cremers
 * Computer Graphics Forum, 2016
 *
 * If you use this code, please cite the paper above.
 * 
 * Emanuele Rodola (c) 2015
 *
 * To compile:
 * mex iidyn.cpp
 */
 
#include <limits>
#include "mex.h"

template <typename T>
inline void mult(const T *A,const T *x, int size, T *y)
{
  const T *ptr;
  double yv;
  for(int j=0;j<size;++j,++y){
	 ptr=x;
	 //*y=0;
	 yv = 0;
	 for(int i=0;i<size;++i,++ptr,++A)
		//*y+=*A**ptr;
		yv += *A * *ptr;
	 *y = static_cast<T>(yv);
  }
}

template <typename T>
inline void simplexify(T *x, int size){
  T sum=0;
  T *ptr=x;
  for(int i=0;i<size;++i,++ptr)
	 if(*ptr>=0)
		sum+=*ptr;
  else
	 *ptr=0;
  for(int i=0;i<size;++i,++x)
	 *x/=sum;
}

template <typename T>
inline double dot(const T *x, const T *y, int size){
  double sum=0;
  for(int i=0;i<size;++i,++x,++y)
	 sum+=*x**y;
  return sum;
}

template <typename T>
inline void scale(T *x, int size, double c){
  for(int i=0;i<size;++i,++x)
	 *x*=c;
}

template <typename T>
inline void linear_comb(const T *x, T *y, int size,double alfa){
  for(int i=0;i<size;++i,++x,++y)
	 *y=alfa*(*x-*y)+*y;
}

template <typename T>
inline double nash_error(const T *x, const T *Ax, double xAx, int size)
{
  double sum=0;
  const T *p_x=x;
  const T *p_Ax=Ax;
  double tmp;
  for(int i=0;i<size;++i,++p_x,++p_Ax){
	 tmp=xAx-*p_Ax;
	 if(tmp>*p_x)
		tmp=*p_x;
	 sum+=tmp*tmp;
  }
  return sum;
}

template <typename T>
inline double selectStrategy(const T *Ax, const T *x, int size, double &delta, int &idx)
{
  const T *ptr1,*ptr2;
  double maxv = -std::numeric_limits<double>::infinity();
  double minv = std::numeric_limits<double>::infinity();

  int max_idx=-1,min_idx=-1;
  ptr1=Ax;
  ptr2=x;

  for(int i=0;i<size;++i,++ptr1,++ptr2){
	 if(*ptr1>maxv){
		maxv=*ptr1;
		max_idx=i;
	 }
	 if(*ptr2>0&&*ptr1<minv){
		minv=*ptr1;
		min_idx=i;
	 }
  }
  double xAx=dot(Ax,x,size);

  maxv-=xAx;
  minv=xAx-minv;

  idx=max_idx;
  delta=maxv;
  if(maxv<minv){
	 idx=min_idx;
	 delta=-minv;
  }

  return nash_error(x,Ax,xAx,size);
}

template <typename T>
void iidyn(const T* A, T* x, int size, T toll, int max_iters, double* out_iters, double* out_err)
{
  int niter=0;
  /* Calculate Ax */
  T *Ax=new T[size];
  simplexify(x,size);
  mult(A,x,size,Ax);

  /*** OCCHIO **/
  toll*=toll;

  double delta = 0.;
  double err = std::numeric_limits<double>::max();

  while (niter < max_iters)
  {
	 int idx=-1;

	 err = selectStrategy(Ax,x,size,delta,idx);
	 if (err < toll)
		break;

	 double den=A[idx*(size+1)]-Ax[idx]-delta;
	 bool do_remove=false;
	 double mu,tmp;
	 if(delta>=0)
	 {
		mu=1;
		if(den<0)
		{
		   tmp=-delta/den;
		   if(mu>tmp)mu=tmp;
		   if(mu<0)mu=0;
		}
	 }
	 else
	 {
		//mu=1.0-1.0/(1-x[idx]);
		mu=x[idx]/(x[idx]-1);
		do_remove=true;
		if(den<0)
		{
		   tmp=-delta/den;
		   if(mu<tmp)
		   {
			  mu=tmp;
			  do_remove=false;
		   }
		   if(mu>0)mu=0;
		}
	 }
	 scale(x,size,1-mu);
	 x[idx]=do_remove?0:x[idx]+mu;

	 simplexify(x,size);

	 linear_comb(A+idx*size, Ax, size, mu);

	 ++niter;
  }
  
  delete[] Ax;
  
  *out_iters = double(niter);
  *out_err = err;

  return;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{ 
    if (nrhs != 4)
		mexErrMsgTxt("Usage: [iters err pop] = iidyn(A, x, toll, max_iters).");
    if (nlhs > 3)
		mexErrMsgTxt("Too many output arguments."); 

	// input

	const double *A = mxGetPr(prhs[0]);	
	
	const int size = int( mxGetN(prhs[1]) );
	
	const double* const xin = mxGetPr(prhs[1]);
	double* x = new double[size];
	for (int i=0; i<size; ++i)
		x[i] = xin[i];
		
	const double toll = *mxGetPr(prhs[2]);
	const int max_iters = int( *mxGetPr(prhs[3]) );
    
	// output
	
	plhs[0] = mxCreateDoubleScalar(0.);
	plhs[1] = mxCreateDoubleScalar(0.);
	double* out_iters = mxGetPr(plhs[0]);
	double* out_err = mxGetPr(plhs[1]);
    
	// run
	
    iidyn(A, x, size, toll, max_iters, out_iters, out_err);
	
	plhs[2] = mxCreateDoubleMatrix(1, size, mxREAL);
	double* out_x = mxGetPr(plhs[2]);
	for (int i=0; i<size; ++i)
		out_x[i] = x[i];
	
	delete[] x;

    return;
}
