/*
 * Implements infection-immunization dynamics by Rota Bulo and Bomze.
 * This is a sparse, optimized version of iidyn().
 * This code makes STRONG ASSUMPTIONS on the input, be careful! 
 *
 *  - Matrix A must have zeros along the diagonal
 *  - A_cols must be sorted increasingly
 *  - A_rows, A_cols and A_cols_lookup must be of int32 type, already 0-based
 *  - init_idx must be of double type, 1-based
 *
 * Compile with: mex -v iidyn_sparse.cpp CXXOPTIMFLAGS="-O2 -fomit-frame-pointer"
 *
 * Written by Emanuele Rodola
 * USI Lugano, July 2016
 */
 
#include <limits>
#include "mex.h"

/*
// Just a sparse version of the original dense function.
// This is not actually ever used in the new sparse code.
template <typename T>
inline void mult_sparse(const int* A_rows, const int* A_cols, const T* A_vals, const T *x, int size, int n_nonzero, T *y)
{
	for (int i=0; i<size; ++i)
		y[i] = 0.;
	for (int k=0; k<n_nonzero; ++k)
		y[A_rows[k]] += A_vals[k]*x[A_cols[k]];
}
*/

template <typename T>
inline void simplexify(T *x, int size){
  T sum=0;
  T *ptr=x;
  for(int i=0;i<size;++i,++ptr)
	 if(*ptr>=0)
		sum+=*ptr;
  else
	 *ptr=0;
  for(int i=0;i<size;++i,++x)
	 *x/=sum;
}

template <typename T>
inline double dot(const T *x, const T *y, int size){
  double sum=0;
  for(int i=0;i<size;++i,++x,++y)
	 sum+=*x**y;
  return sum;
}

template <typename T>
inline void scale(T *x, int size, double c){
  for(int i=0;i<size;++i,++x)
	 *x*=c;
}

/*
template <typename T>
inline void linear_comb_sparse(const int* A_rows, const T* A_vals, int start, int len, T *y, int size, double alfa)
{
	for (int i=0; i<size; ++i)
		y[i] -= alfa*y[i];
	for (int i=start; i<start+len; ++i)
		y[A_rows[i]] += alfa*A_vals[i];
}
*/

template <typename T>
inline void linear_comb_sparse(const int* A_rows, const T* A_vals, int start, const int* A_cols, T *y, int size, double alfa)
{
	for (int i=0; i<size; ++i)
		y[i] -= alfa*y[i];
	
	const int v = A_cols[start];
	for (int i=start; i<start+size; ++i)
	{
		if (A_cols[i] != v)
			break;
		y[A_rows[i]] += alfa*A_vals[i];
	}
}

template <typename T>
inline double nash_error(const T *x, const T *Ax, double xAx, int size)
{
  double sum=0;
  const T *p_x=x;
  const T *p_Ax=Ax;
  double tmp;
  for(int i=0;i<size;++i,++p_x,++p_Ax){
	 tmp=xAx-*p_Ax;
	 if(tmp>*p_x)
		tmp=*p_x;
	 sum+=tmp*tmp;
  }
  return sum;
}

template <typename T>
inline double selectStrategy(const T *Ax, const T *x, int size, double &delta, int &idx, double* xAx)
{
  const T *ptr1,*ptr2;
  double maxv = -std::numeric_limits<double>::infinity();
  double minv = std::numeric_limits<double>::infinity();

  int max_idx=-1,min_idx=-1;
  ptr1=Ax;
  ptr2=x;

  for(int i=0;i<size;++i,++ptr1,++ptr2){
	 if(*ptr1>maxv){
		maxv=*ptr1;
		max_idx=i;
	 }
	 if(*ptr2>0&&*ptr1<minv){
		minv=*ptr1;
		min_idx=i;
	 }
  }
  *xAx = dot(Ax,x,size);

  maxv -= *xAx;
  minv = *xAx - minv;

  idx=max_idx;
  delta=maxv;
  if(maxv<minv){
	 idx=min_idx;
	 delta=-minv;
  }

  return nash_error(x,Ax,*xAx,size);
}

template <typename T>
void iidyn(const int* A_rows, const int* A_cols, const T* A_vals, const int* A_cols_lookup, int init_idx, T* x, int size, int n_nonzero, T toll, int max_iters, double* out_iters, double* out_err, double* energy)
{
  T *Ax = new T[size];
  
  for (int i=0; i<size; ++i)
	  Ax[i] = 0.;
  for (int i=A_cols_lookup[init_idx]; i<n_nonzero; ++i)
  {
	  if (A_cols[i] != init_idx)
		  break;
	  Ax[A_rows[i]] = A_vals[i];
  }
  
  /*
  // this commented block of code is computing the A_cols_lookup and len
  // structures, in case for some crazy reasons these can't be
  // passed directly from matlab.
  
  int* A_cols_lookup = new int[size];
  
  for (int i=0; i<size; ++i)
  {
	  Ax[i] = 0.;
	  A_cols_lookup[i] = -1;
  }
	  
  int* len = new int[size];
  int newidx;
  for (int i=0; i<n_nonzero; ++i)
  {
	  newidx = A_cols[i];
	  if (A_cols_lookup[newidx]<0)
	  {
		  A_cols_lookup[newidx] = i;
		  len[newidx] = 1;
	  }
	  else
	  {
		  len[newidx] += 1;
	  }
	  // just the init_idx-th column of A
	  if (newidx==init_idx)
		  Ax[A_rows[i]] = A_vals[i];
  }*/
  
  //mult_sparse(A_rows, A_cols, A_vals, x, size, n_nonzero, Ax);

  toll*=toll;

  double delta = 0.;
  double err = std::numeric_limits<double>::max();
  
  int niter=0;
  while (niter < max_iters)
  {
	 int idx=-1;

	 err = selectStrategy(Ax,x,size,delta,idx,energy);
	 if (err < toll)
		break;

	 // note that A[idx*(size+1)] is just the idx-th value along the diagonal
	 // double den = A[idx*(size+1)]-Ax[idx]-delta;
	 
	 double den = -Ax[idx]-delta;
	 
	 bool do_remove=false;
	 double mu,tmp;
	 if(delta>=0)
	 {
		mu=1;
		if(den<0)
		{
		   tmp=-delta/den;
		   if(mu>tmp)mu=tmp;
		   if(mu<0)mu=0;
		}
	 }
	 else
	 {
		mu=x[idx]/(x[idx]-1);
		do_remove=true;
		if(den<0)
		{
		   tmp=-delta/den;
		   if(mu<tmp)
		   {
			  mu=tmp;
			  do_remove=false;
		   }
		   if(mu>0)mu=0;
		}
	 }
	 scale(x,size,1-mu);
	 x[idx]=do_remove?0:x[idx]+mu;

	 simplexify(x,size);
	 
	 //linear_comb_sparse(A_rows, A_vals, A_cols_lookup[idx], len[idx], Ax, size, mu);
	 linear_comb_sparse(A_rows, A_vals, A_cols_lookup[idx], A_cols, Ax, size, mu);

	 ++niter;
  }
  
  delete[] Ax;

  *out_iters = double(niter);
  *out_err = err;

  return;
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
    if (nrhs != 8)
		mexErrMsgTxt("Usage: [iters, err, x, energy] = iidyn_sparse(A_rows, A_cols, A_vals, A_cols_lookup, size, init_idx, toll, max_iters).");
    if (nlhs > 4)
		mexErrMsgTxt("Too many output arguments."); 

	// sparse representation of the similarity matrix A
	const int* A_rows = (int*)mxGetData(prhs[0]);
	const int* A_cols = (int*)mxGetData(prhs[1]);
	const double *A_vals = mxGetPr(prhs[2]);
	const int* A_cols_lookup = (int*)mxGetData(prhs[3]);
	
	const int size = int(*mxGetPr(prhs[4]));
	const int n_nonzero = int( mxGetN(prhs[0])>mxGetM(prhs[0])?mxGetN(prhs[0]):mxGetM(prhs[0]) );
	
	//mexPrintf("Similarity matrix is %dx%d with %d non-zero elements.\n", size, size, n_nonzero);
	
	// the vertex on the simplex where the optimization starts
	const int init_idx = int(*mxGetPr(prhs[5]))-1;
		
	const double toll = *mxGetPr(prhs[6]);
	const int max_iters = int( *mxGetPr(prhs[7]) );
	
	plhs[0] = mxCreateDoubleScalar(0.);
	plhs[1] = mxCreateDoubleScalar(0.);
	plhs[3] = mxCreateDoubleScalar(0.);
	double* out_iters = mxGetPr(plhs[0]);
	double* out_err = mxGetPr(plhs[1]);
    double* out_energy = mxGetPr(plhs[3]);
	
	plhs[2] = mxCreateDoubleMatrix(1, size, mxREAL);
	double* out_x = mxGetPr(plhs[2]);
	
	for (int i=0; i<size; ++i)
		out_x[i] = (i==init_idx?1.:0.);
	
    iidyn(A_rows, A_cols, A_vals, A_cols_lookup, init_idx, out_x, size, n_nonzero, toll, max_iters, out_iters, out_err, out_energy);

    return;
}
