%
% This demo implements the method described in:
%
% "A Game-Theoretic Approach to Deformable Shape Matching"
% Rodola, Bronstein, Albarelli, Bergamasco, Torsello
% Proc. CVPR 2012
%
% Please reference the paper above if you use this in your research.
%
function result = match_qap_l1(M, S, options)

result = struct;

if nargin<3
    options = struct;
end

if ~isfield(options, 'tmp_dir')
    options.tmp_dir = './tmp/';
end
if ~isfield(options, 'densify')
    options.densify = false;
end
if ~isfield(options, 'densify_amount')
    options.densify_amount = 0.5; % this is the alpha in elastic net
end
if ~isfield(options, 'n_samples')
    options.n_samples = 1e3;
end
if ~isfield(options, 'min_boundary_dist')
    options.min_boundary_dist = 5; % percent of model diameter
end
if ~isfield(options, 'descr_knn')
    options.descr_knn = 10;
end
if ~isfield(options, 'descr_dist_thresh')
    options.descr_dist_thresh = Inf; 
end
if ~isfield(options, 'multi_start')
    options.multi_start = false;
end
if ~isfield(options, 'rigid')
    options.rigid = false;
end

if options.densify && (options.densify_amount<=0 || options.densify_amount>=1)
    error('Elastic net requires alpha values strictly in (0,1). Provided: %.2f.', options.densify_amount)
end

if ~exist(options.tmp_dir, 'dir')
    mkdir(options.tmp_dir);
end

% -------------------------------------------------------------------------
% Sample model and scene
% -------------------------------------------------------------------------

diam = sqrt(sum(calc_tri_areas(M))); % this approximates geodesic diameter
max_radius = diam * options.min_boundary_dist/100;

tmp_fname = sprintf('%s/%s.mat', options.tmp_dir, M.name);
if ~exist(tmp_fname, 'file')
    model_samples = fps_euclidean(M.VERT, options.n_samples, 1);
    DM = calc_dist_matrix(M, model_samples);
    save(tmp_fname, 'model_samples', 'DM');
else
    load(tmp_fname);
    assert(length(model_samples)==options.n_samples)
end

tmp_fname = sprintf('%s/%s.mat', options.tmp_dir, S.name);
if ~exist(tmp_fname, 'file')
    scene_samples = fps_euclidean(S.VERT, options.n_samples, 1);
    scene_samples = remove_boundary_samples(S, scene_samples, max_radius);
    DS = calc_dist_matrix(S, scene_samples);
    precomputed_max_radius = max_radius; % just for sanity checks
    save(tmp_fname, 'scene_samples', 'DS', 'precomputed_max_radius');
else
    load(tmp_fname);
    assert(precomputed_max_radius == max_radius)
    % if the assertion fails, the precomputed data was obtained with a
    % different radius
end

if options.rigid % just overwrite geodesics with Euclidean
    DM = pdist2(M.VERT(model_samples,:),M.VERT(model_samples,:));
    DS = pdist2(S.VERT(scene_samples,:),S.VERT(scene_samples,:));
end

M.desc = M.desc(model_samples,:);
S.desc = S.desc(scene_samples,:);

% -------------------------------------------------------------------------
% Generate candidate matches
% -------------------------------------------------------------------------

tic

fprintf('Generating candidate matches... ');

% ann_params = struct;
% ann_params.algorithm = 'linear';
% ann_params.trees = 8;
% ann_params.checks = 64;
% ann_params.centers_init = 'kmeanspp';
% ann_params.iterations = -1;
% [matches_knn, dists_knn] = flann_search(M.desc', S.desc', options.descr_knn, ann_params);
% dists_knn = sqrt(dists_knn);

[matches_knn, dists_knn] = knnsearch(M.desc, S.desc, 'K', options.descr_knn);
matches_knn = matches_knn';
dists_knn = dists_knn';

% matches are (scene,model) pairs
idx_to_keep = dists_knn < options.descr_dist_thresh;
[~,idx] = find(idx_to_keep);
candidate_matches = [idx , matches_knn(idx_to_keep)];

fprintf('done.\n')
fprintf('%d candidate matches.\n', size(candidate_matches,1));

result.candidate_matches = candidate_matches;

% -------------------------------------------------------------------------
% Construct similarity matrix
% -------------------------------------------------------------------------

fprintf('Constructing similarity matrix...\n');

% it will be ensured that no matches will be less than min_matches_dist apart
min_matches_dist = diam*options.min_matches_dist;

sim = calc_sim_matrix_wellspread(...
    candidate_matches', ...
    DS, ...
    DM, ...
    1e-2, ...
    S.desc', ...
    M.desc', ...
    min_matches_dist);

fprintf('done.\n');

% -------------------------------------------------------------------------
% Solve L1-relaxed QAP
% -------------------------------------------------------------------------

if options.multi_start
    
    n_runs = size(candidate_matches,1);
    runs = cell(1,n_runs);
    energies = zeros(1,n_runs);
    
    % we will work with a sparse similarity matrix
    sim = sparse(sim);
    [sim_rows, sim_cols, sim_vals] = find(sim);
    sim_rows = int32(sim_rows)-1;
    sim_cols = int32(sim_cols)-1;
    [~,sim_cols_lookup] = ismember((1:size(sim,1))-1, sim_cols);
    sim_cols_lookup = int32(sim_cols_lookup)-1;
    
    assert(all(sort(sim_cols) == sim_cols));

    fprintf('Optimizing multi-start with %d initial points...\n', n_runs);

    parfor i=1:n_runs
        runs{i} = struct;
        [runs{i}.iters, runs{i}.err, runs{i}.sol, energies(i)] = ...
            iidyn_sparse(...
            sim_rows, sim_cols, sim_vals, sim_cols_lookup, size(sim,1),...
            i, 1e-5, 1e5);
    end
    
    % keep the solution with largest total similarity
    [final_energy, best_run] = max(energies);
    sol = runs{best_run}.sol;
    err = runs{best_run}.err;
    iters = runs{best_run}.iters;
    
else % start from the (perturbed) barycenter
    
    fprintf('Single start optimization (barycenter)... ');
    
    xinit = ones(1,size(candidate_matches,1))./size(candidate_matches,1);
    xinit = xinit + rand(size(xinit));
    xinit = xinit ./ sum(xinit);
    [iters, err, sol] = iidyn(sim, xinit, 1e-6, 1e5);
    
    final_energy = sol*sim*sol';
    
    fprintf('done.\n')
    
end

final_matches = candidate_matches(sol>0,:);

result.elapsed_time = toc;

fprintf('%d iterations at convergence\n', iters)
fprintf('Error at convergence: %.6f\n', err)
fprintf('%d final matches\n', size(final_matches,1))
fprintf('Final energy (max is 1.0): %.4f\n', final_energy)

fprintf('Elapsed time: %.2f sec.\n', result.elapsed_time);

result.matches = [scene_samples(final_matches(:,1)), model_samples(final_matches(:,2))];
result.weights = sol(sol>0)';
result.final_energy = final_energy;

% -------------------------------------------------------------------------
% Optionally densify using elastic net
% -------------------------------------------------------------------------

if options.densify
    
    fprintf('Densifying...\n')
    
    tic;
    
    [x_enet, iters_enet, err_enet] = ...
        enet_grad_desc(sim, options.densify_amount, sol, 1e-5, 1e4, 1e-4);
    
    result.elapsed_time_densify = toc;
    
    fprintf('%d iterations at convergence\n', iters_enet)
    fprintf('Error at convergence: %.6f\n', err_enet)
    
    fprintf('Elapsed time of densification step: %.2f sec.\n', result.elapsed_time_densify);
    
    x_enet(x_enet<0.05*max(x_enet)) = 0;
    matches = prune_matches(x_enet, double(candidate_matches-1)');
    
    final_matches = candidate_matches(matches,:);
    
    fprintf('%d matches after densification\n', size(final_matches,1))
    
    result.matches_densified = [scene_samples(final_matches(:,1)), model_samples(final_matches(:,2))];
    result.weights_densified = x_enet(matches);
    
end

end
